import httpx
import json
from typing import Any, AsyncGenerator, Type, TypeVar
from dialtone.errors import (
    APIErrorRouterDetails,
    BadRequestError,
    AuthenticationError,
    NotFoundError,
    UnprocessableEntityError,
    RateLimitError,
    InternalServerError,
    BadGatewayError,
    ServiceUnavailableError,
    ProviderModerationError,
    ConfigurationError,
    APIError,
    ErrorCode,
    StatusCode,
)
from dialtone.types import ChatCompletionChunk
from dialtone.config import DEFAULT_REQUEST_TIMEOUT

# Generic type variable
T = TypeVar("T")


def get_status_code_from_error_code(error_code: ErrorCode) -> StatusCode:
    STATUS_CODE_FROM_CODE = {
        ErrorCode.bad_request: StatusCode.bad_request,
        ErrorCode.unauthorized: StatusCode.unauthorized,
        ErrorCode.not_found: StatusCode.not_found,
        ErrorCode.unprocessable_entity: StatusCode.unprocessable_entity,
        ErrorCode.too_many_requests: StatusCode.too_many_requests,
        ErrorCode.internal_server_error: StatusCode.internal_server_error,
        ErrorCode.bad_gateway: StatusCode.bad_gateway,
        ErrorCode.service_unavailable: StatusCode.service_unavailable,
    }

    return STATUS_CODE_FROM_CODE[error_code]


def get_error_code_from_status_code(status_code: StatusCode) -> ErrorCode:
    CODE_FROM_STATUS_CODE = {
        StatusCode.bad_request: ErrorCode.bad_request,
        StatusCode.unauthorized: ErrorCode.unauthorized,
        StatusCode.not_found: ErrorCode.not_found,
        StatusCode.unprocessable_entity: ErrorCode.unprocessable_entity,
        StatusCode.too_many_requests: ErrorCode.too_many_requests,
        StatusCode.internal_server_error: ErrorCode.internal_server_error,
        StatusCode.bad_gateway: ErrorCode.bad_gateway,
        StatusCode.service_unavailable: ErrorCode.service_unavailable,
    }

    return CODE_FROM_STATUS_CODE[status_code]


def get_error_class_from_status_code(status_code: StatusCode) -> Type[APIError]:
    ERROR_CLASS_FROM_STATUS_CODE = {
        StatusCode.bad_request: BadRequestError,
        StatusCode.unauthorized: AuthenticationError,
        StatusCode.not_found: NotFoundError,
        StatusCode.unprocessable_entity: UnprocessableEntityError,
        StatusCode.too_many_requests: RateLimitError,
        StatusCode.internal_server_error: InternalServerError,
        StatusCode.bad_gateway: BadGatewayError,
        StatusCode.service_unavailable: ServiceUnavailableError,
    }

    return ERROR_CLASS_FROM_STATUS_CODE[status_code]


def get_error_class_from_error_code(error_code: ErrorCode) -> Type[APIError]:
    ERROR_CLASS_FROM_ERROR_CODE = {
        ErrorCode.bad_request: BadRequestError,
        ErrorCode.unauthorized: AuthenticationError,
        ErrorCode.not_found: NotFoundError,
        ErrorCode.unprocessable_entity: UnprocessableEntityError,
        ErrorCode.too_many_requests: RateLimitError,
        ErrorCode.internal_server_error: InternalServerError,
        ErrorCode.bad_gateway: BadGatewayError,
        ErrorCode.service_unavailable: ServiceUnavailableError,
        ErrorCode.provider_moderation: ProviderModerationError,
        ErrorCode.configuration_error: ConfigurationError,
    }

    return ERROR_CLASS_FROM_ERROR_CODE[error_code]


def process_response(response: httpx.Response):
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError as exc:
        error_params = {
            "status_code": StatusCode(exc.response.status_code),
            "request": exc.request,
            "response": exc.response,
        }

        try:
            response_body: dict = exc.response.json()

            error_code = response_body.get("detail", {}).get("error_code")

            if error_code:
                # Happy path
                error_class = get_error_class_from_error_code(ErrorCode(error_code))

                if response_body.get("detail", {}).get("message"):
                    error_params["message"] = response_body["detail"]["message"]

                if response_body.get("detail", {}).get("router_details"):
                    error_params["router_details"] = APIErrorRouterDetails(
                        **response_body["detail"]["router_details"]
                    )

                raise error_class(**error_params) from None
            else:
                # Response body is JSON but is invalid format.
                error_class = get_error_class_from_status_code(
                    error_params["status_code"]
                )
                raise error_class(**error_params) from None
        except json.decoder.JSONDecodeError:
            # Response body is not valid JSON
            error_class = get_error_class_from_status_code(error_params["status_code"])
            if exc.response.text:
                error_params["message"] = exc.response.text

            raise error_class(**error_params) from None

    return response.json()


def dialtone_post_request(
    url: str,
    data: dict[str, Any],
    headers: dict[str, str],
    timeout: int = DEFAULT_REQUEST_TIMEOUT,
) -> dict:
    with httpx.Client() as client:
        response = client.post(url, json=data, headers=headers, timeout=timeout)
        return process_response(response)


async def dialtone_post_request_async(
    url: str,
    data: dict[str, Any],
    headers: dict[str, str],
    timeout: int = DEFAULT_REQUEST_TIMEOUT,
) -> dict:
    async with httpx.AsyncClient() as client:
        response = await client.post(url, json=data, headers=headers, timeout=timeout)
        return process_response(response)


async def dialtone_streaming_post_request_async(
    url: str,
    data: dict[str, Any],
    headers: dict[str, str],
    timeout: int = DEFAULT_REQUEST_TIMEOUT,
) -> AsyncGenerator[dict, None]:
    async with httpx.AsyncClient() as client:
        async with client.stream(
            "POST", url, json=data, headers=headers, timeout=timeout
        ) as response:
            async for response_chunk_text in response.aiter_text():
                response_chunk_json: dict = json.loads(
                    response_chunk_text.removeprefix("data:")
                )
                yield response_chunk_json


async def convert_dict_to_type_stream(
    generator: AsyncGenerator[dict, None], converter_type: Type[T]
) -> AsyncGenerator[T, None]:
    async for item in generator:
        yield converter_type(**item)



import httpx
from enum import Enum
from pydantic import BaseModel
from dialtone.types import LLM, Provider


class ErrorCode(Enum):
    # Standard
    bad_request = "bad_request"
    unauthorized = "unauthorized"
    not_found = "not_found"
    unprocessable_entity = "unprocessable_entity"
    too_many_requests = "too_many_requests"
    internal_server_error = "internal_server_error"
    bad_gateway = "bad_gateway"
    service_unavailable = "service_unavailable"
    # Custom
    provider_moderation = "provider_moderation"
    configuration_error = "configuration_error"

    def __str__(self):
        return str(self.value)


class StatusCode(Enum):
    bad_request = 400
    unauthorized = 401
    not_found = 404
    unprocessable_entity = 422
    too_many_requests = 429
    internal_server_error = 500
    bad_gateway = 502
    service_unavailable = 503

    def __str__(self):
        return str(self.value)


class DialtoneError(Exception):
    pass


class APIErrorRouterDetails(BaseModel):
    model: LLM | None = None
    provider: Provider | None = None
    provider_response: dict | None = None


class APIError(DialtoneError):
    request: httpx.Request
    message: str
    router_details: APIErrorRouterDetails

    def __str__(self):
        if self.router_details.provider_response:
            return (
                f"{self.message} (model {self.router_details.model} from "
                f"{self.router_details.provider}) - Provider Response: {self.router_details.provider_response}"
            )
        return self.message


class APIStatusError(APIError):
    response: httpx.Response
    status_code: StatusCode

    def __init__(
        self,
        request: httpx.Request,
        response: httpx.Response,
        status_code: StatusCode | None = None,
        message: str | None = None,
        router_details: APIErrorRouterDetails = APIErrorRouterDetails(),
    ):
        self.request = request
        self.response = response
        self.router_details = router_details

        if status_code:
            self.status_code = status_code
        if message:
            self.message = message


class BadRequestError(APIStatusError):
    status_code: StatusCode = StatusCode.bad_request
    message: str = "Bad Request"


class AuthenticationError(APIStatusError):
    status_code: StatusCode = StatusCode.unauthorized
    message: str = "Unauthorized"


class NotFoundError(APIStatusError):
    status_code: StatusCode = StatusCode.not_found
    message: str = "Not Found"


class UnprocessableEntityError(APIStatusError):
    status_code: StatusCode = StatusCode.unprocessable_entity
    message: str = "Unprocessable Entity"


class RateLimitError(APIStatusError):
    status_code: StatusCode = StatusCode.too_many_requests
    message: str = "Too Many Requests"


class InternalServerError(APIStatusError):
    status_code: StatusCode = StatusCode.internal_server_error
    message: str = "Internal Server Error"


class BadGatewayError(APIStatusError):
    status_code: StatusCode = StatusCode.bad_gateway
    message: str = "Bad Gateway"


class ServiceUnavailableError(APIStatusError):
    status_code: StatusCode = StatusCode.service_unavailable
    message: str = "Service Unavailable"


class ProviderModerationError(APIStatusError):
    status_code: StatusCode = StatusCode.bad_request
    message: str = "Provider Moderation Error"


class ConfigurationError(APIStatusError):
    status_code: StatusCode = StatusCode.bad_request
    message: str = "Configuration Error"



from typing import Any
from pydantic import BaseModel
from dialtone.types import (
    FallbackConfig,
    ProviderConfig,
    RouterModelConfig,
    ChatCompletion,
    Choice,
    ChatMessage,
    Tool,
    DialtoneClient,
    Dials,
    RouteDecision,
    TokenUsage,
    ToolsConfig,
)
from dialtone.dialtone.dialtone_base import DialtoneBase
from dialtone.utils.api import dialtone_post_request
from dialtone.utils.prepare_payload import prepare_chat_completion, prepare_chat_route
from dialtone.config import DEFAULT_BASE_URL


class Completions(BaseModel):
    client: DialtoneClient

    def create(
        self, messages: list[ChatMessage] | list[dict[str, Any]], tools: list[Tool] = []
    ):
        headers, params = prepare_chat_completion(
            messages=messages, tools=tools, client=self.client
        )

        response_json = dialtone_post_request(
            url=f"{self.client.base_url}/chat/completions",
            data=params,
            headers=headers,
        )

        return ChatCompletion(
            model=response_json["model"],
            provider=response_json["provider"],
            choices=[
                Choice(
                    message=response_json["message"],
                )
            ],
            usage=TokenUsage(**response_json["usage"]),
        )


class Chat(BaseModel):
    client: DialtoneClient
    completions: Completions

    def __init__(self, client: DialtoneClient):
        completions = Completions(client=client)
        super().__init__(client=client, completions=completions)

    def route(
        self, messages: list[ChatMessage] | list[dict[str, Any]], tools: list[Tool] = []
    ):
        headers, params = prepare_chat_route(
            messages=messages, tools=tools, client=self.client
        )

        response_json = dialtone_post_request(
            url=f"{self.client.base_url}/chat/route",
            data=params,
            headers=headers,
            timeout=15,
        )

        return RouteDecision(
            model=response_json["model"],
            providers=response_json["providers"],
            quality_predictions=response_json["quality_predictions"],
            routing_strategy=response_json["routing_strategy"],
        )


class Dialtone(DialtoneBase):
    chat: Chat
    client: DialtoneClient

    def __init__(
        self,
        api_key: str,
        provider_config: ProviderConfig | dict[str, Any],
        dials: Dials | dict[str, Any] = Dials(),
        router_model_config: RouterModelConfig | dict[str, Any] = RouterModelConfig(),
        fallback_config: FallbackConfig | dict[str, Any] = FallbackConfig(),
        tools_config: ToolsConfig | dict[str, Any] = ToolsConfig(),
        base_url: str = DEFAULT_BASE_URL,
    ):
        super().validate_inputs(
            provider_config=provider_config,
            dials=dials,
            router_model_config=router_model_config,
            fallback_config=fallback_config,
            tools_config=tools_config,
        )

        if isinstance(dials, dict):
            dials = Dials(**dials)
        if isinstance(provider_config, dict):
            provider_config = ProviderConfig(**provider_config)
        if isinstance(router_model_config, dict):
            router_model_config = RouterModelConfig(**router_model_config)
        if isinstance(fallback_config, dict):
            fallback_config = FallbackConfig(**fallback_config)
        if isinstance(tools_config, dict):
            tools_config = ToolsConfig(**tools_config)

        self.client = DialtoneClient(
            api_key=api_key,
            dials=dials,
            provider_config=provider_config,
            router_model_config=router_model_config,
            fallback_config=fallback_config,
            tools_config=tools_config,
            base_url=base_url,
        )
        self.chat = Chat(client=self.client)



from typing import Any
from dialtone.types import ChatMessage, Tool, DialtoneClient


def prepare_chat_message(message: ChatMessage | dict[str, Any]) -> dict:
    if isinstance(message, dict):
        return message

    result: dict[str, Any] = {
        "role": message.role,
        "content": message.content,
    }
    if message.tool_calls:
        result["tool_calls"] = (
            [
                {
                    "id": tool_call.id,
                    "type": tool_call.type,
                    "function": {
                        "name": (
                            tool_call.function.name if tool_call.function else None
                        ),
                        "arguments": (
                            tool_call.function.arguments if tool_call.function else None
                        ),
                    },
                }
                for tool_call in message.tool_calls
            ]
            if message.tool_calls
            else []
        )
    return result


def prepare_chat_completion(
    client: DialtoneClient,
    messages: list[ChatMessage] | list[dict[str, Any]],
    stream: bool = False,
    tools: list[Tool] = [],
) -> tuple[dict, dict]:
    headers = {"Authorization": f"Bearer {client.api_key}"}
    params = {
        "messages": [prepare_chat_message(message) for message in messages],
        "dials": client.dials.model_dump(),
        "provider_config": client.provider_config.model_dump(),
    }
    if stream:
        params["stream"] = True
    if client.router_model_config:
        params["router_model_config"] = client.router_model_config.model_dump()
    if client.fallback_config:
        params["fallback_config"] = client.fallback_config.model_dump()
    if client.tools_config:
        params["tools_config"] = client.tools_config.model_dump()
    if tools:
        params["tools"] = [tool.model_dump() for tool in tools]

    return headers, params


def prepare_chat_route(
    client: DialtoneClient,
    messages: list[ChatMessage] | list[dict[str, Any]],
    tools: list[Tool] = [],
) -> tuple[dict, dict]:
    headers = {"Authorization": f"Bearer {client.api_key}"}
    params = {
        "messages": [prepare_chat_message(message) for message in messages],
        "dials": client.dials.model_dump(),
        "provider_config": client.provider_config.model_dump(),
    }
    if client.router_model_config:
        params["router_model_config"] = client.router_model_config.model_dump()
    if client.fallback_config:
        params["fallback_config"] = client.fallback_config.model_dump()
    if client.tools_config:
        params["tools_config"] = client.tools_config.model_dump()
    if tools:
        params["tools"] = [tool.model_dump() for tool in tools]

    return headers, params



from typing import Any
from pydantic import ValidationError
from dialtone.types import (
    ProviderConfig,
    Dials,
    RouterModelConfig,
    FallbackConfig,
    ToolsConfig,
    LLM,
)


class DialtoneBase:
    def validate_inputs(
        self,
        provider_config: ProviderConfig | dict[str, Any],
        dials: Dials | dict[str, Any] = Dials(),
        router_model_config: RouterModelConfig | dict[str, Any] = RouterModelConfig(),
        fallback_config: FallbackConfig | dict[str, Any] = FallbackConfig(),
        tools_config: ToolsConfig | dict[str, Any] = ToolsConfig(),
    ):
        try:
            if isinstance(provider_config, dict):
                provider_config = ProviderConfig(**provider_config)
        except ValidationError as e:
            raise ValidationError(f"Invalid provider_config: {e}")

        try:
            if isinstance(dials, dict):
                dials = Dials(**dials)
        except ValidationError as e:
            raise ValidationError(f"Invalid dials: {e}")

        try:
            if isinstance(router_model_config, dict):
                router_model_config = RouterModelConfig(**router_model_config)
        except ValidationError as e:
            raise ValidationError(f"Invalid router_model_config: {e}")

        try:
            if router_model_config.include_models:
                router_model_config.include_models = [
                    LLM(model) if isinstance(model, str) else model
                    for model in router_model_config.include_models
                ]
        except ValueError as e:
            raise ValidationError(f"Invalid include_models: {e}")

        try:
            if router_model_config.exclude_models:
                router_model_config.exclude_models = [
                    LLM(model) if isinstance(model, str) else model
                    for model in router_model_config.exclude_models
                ]
        except ValueError as e:
            raise ValidationError(f"Invalid exclude_models: {e}")

        try:
            if isinstance(fallback_config, dict):
                fallback_config = FallbackConfig(**fallback_config)
        except ValidationError as e:
            raise ValidationError(f"Invalid fallback_config: {e}")

        try:
            if isinstance(tools_config, dict):
                tools_config = ToolsConfig(**tools_config)
        except ValidationError as e:
            raise ValidationError(f"Invalid tools_config: {e}")



import sys
from enum import StrEnum
from pydantic import BaseModel
from typing import List, Literal, Any, Optional
from dialtone.config import DEFAULT_BASE_URL


class Tool(BaseModel):
    type: Literal["function"]
    function: dict[str, Any]


class ToolCallFunction(BaseModel):
    name: str
    arguments: str


class ChoiceDeltaToolCallFunction(BaseModel):
    name: Optional[str] = None
    arguments: Optional[str] = None


class ToolCall(BaseModel):
    id: str
    type: Literal["function"]
    function: ToolCallFunction


class ChoiceDeltaToolCall(BaseModel):
    index: int
    id: Optional[str] = None
    type: Optional[Literal["function"]] = None
    function: Optional[ChoiceDeltaToolCallFunction] = None


class ChatMessage(BaseModel):
    role: Literal["system", "user", "assistant", "tool"]
    content: str
    tool_calls: list[ToolCall] = []
    tool_call_id: Optional[str] = None
    name: Optional[str] = None


class ChoiceDelta(BaseModel):
    role: Optional[Literal["system", "user", "assistant", "tool"]] = None
    content: Optional[str] = None
    tool_calls: Optional[List[ChoiceDeltaToolCall]] = None


class Provider(StrEnum):
    OpenAI = "openai"
    Groq = "groq"
    DeepInfra = "deepinfra"
    Fireworks = "fireworks"
    Together = "together"
    Replicate = "replicate"
    Anthropic = "anthropic"
    Google = "google"
    Cohere = "cohere"

    def __str__(self):
        return self.value


class LLM(StrEnum):
    claude_3_5_sonnet = "claude-3-5-sonnet-20240620"
    claude_3_haiku = "claude-3-haiku-20240307"
    gpt_4o = "gpt-4o-2024-05-13"
    gpt_4o_mini = "gpt-4o-mini-2024-07-18"
    gemini_1_5_pro = "gemini-1.5-pro"
    gemini_1_5_flash = "gemini-1.5-flash"
    command_r_plus = "command-r-plus"
    command_r = "command-r"
    llama_3_70b = "llama3-70b-8192"

    def __str__(self):
        return self.value


class Choice(BaseModel):
    message: ChatMessage


class ChunkChoice(BaseModel):
    delta: ChoiceDelta
    finish_reason: Optional[
        Literal["stop", "length", "tool_calls", "content_filter", "function_call"]
    ] = None


class TokenUsage(BaseModel):
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0


class ChatCompletion(BaseModel):
    choices: list[Choice]
    model: LLM
    provider: Provider
    usage: TokenUsage


class ChatCompletionChunk(BaseModel):
    model: LLM | None
    provider: Provider | None
    choices: list[ChunkChoice]
    usage: TokenUsage | None


class OpenAIProviderConfig(BaseModel):
    api_key: str


class AnthropicProviderConfig(BaseModel):
    api_key: str


class GoogleProviderConfig(BaseModel):
    api_key: str


class CohereProviderConfig(BaseModel):
    api_key: str


class GroqProviderConfig(BaseModel):
    api_key: str


class ReplicateProviderConfig(BaseModel):
    api_key: str


class FireworksProviderConfig(BaseModel):
    api_key: str


class TogetherProviderConfig(BaseModel):
    api_key: str


class DeepInfraProviderConfig(BaseModel):
    api_key: str


CohereProviders = [Provider.Cohere]

OpenAiProviders = [Provider.OpenAI]

Llama370bNoToolsProviders = [
    Provider.Groq,
    Provider.DeepInfra,
    Provider.Fireworks,
    Provider.Together,
    Provider.Replicate,
]

Llama370bToolsProviders = [Provider.Groq, Provider.DeepInfra]

AnthropicProviders = [Provider.Anthropic]

GoogleProviders = [Provider.Google]


class OpenAIModelConfig(BaseModel):
    providers: list[Provider] = OpenAiProviders


class AnthropicModelConfig(BaseModel):
    providers: list[Provider] = AnthropicProviders


class GoogleModelConfig(BaseModel):
    providers: list[Provider] = GoogleProviders


class CohereModelConfig(BaseModel):
    providers: list[Provider] = CohereProviders


class Llama370BModelConfig(BaseModel):
    tools_providers: list[Provider] = Llama370bToolsProviders
    no_tools_providers: list[Provider] = Llama370bNoToolsProviders


class ProviderConfig(BaseModel):
    openai: Optional[OpenAIProviderConfig] = None
    anthropic: Optional[AnthropicProviderConfig] = None
    google: Optional[GoogleProviderConfig] = None
    cohere: Optional[CohereProviderConfig] = None
    groq: Optional[GroqProviderConfig] = None
    replicate: Optional[ReplicateProviderConfig] = None
    fireworks: Optional[FireworksProviderConfig] = None
    together: Optional[TogetherProviderConfig] = None
    deepinfra: Optional[DeepInfraProviderConfig] = None

    @classmethod
    def OpenAI(cls, api_key: str) -> OpenAIProviderConfig:
        return OpenAIProviderConfig(api_key=api_key)

    @classmethod
    def Anthropic(cls, api_key: str) -> AnthropicProviderConfig:
        return AnthropicProviderConfig(api_key=api_key)

    @classmethod
    def Google(cls, api_key: str) -> GoogleProviderConfig:
        return GoogleProviderConfig(api_key=api_key)

    @classmethod
    def Cohere(cls, api_key: str) -> CohereProviderConfig:
        return CohereProviderConfig(api_key=api_key)

    @classmethod
    def Groq(cls, api_key: str) -> GroqProviderConfig:
        return GroqProviderConfig(api_key=api_key)

    @classmethod
    def Replicate(cls, api_key: str) -> ReplicateProviderConfig:
        return ReplicateProviderConfig(api_key=api_key)

    @classmethod
    def Fireworks(cls, api_key: str) -> FireworksProviderConfig:
        return FireworksProviderConfig(api_key=api_key)

    @classmethod
    def Together(cls, api_key: str) -> TogetherProviderConfig:
        return TogetherProviderConfig(api_key=api_key)

    @classmethod
    def DeepInfra(cls, api_key: str) -> DeepInfraProviderConfig:
        return DeepInfraProviderConfig(api_key=api_key)


class RouterModelConfig(BaseModel):
    include_models: list[LLM | str] = []
    exclude_models: list[LLM | str] = []

    gpt_4o: OpenAIModelConfig = OpenAIModelConfig(providers=OpenAiProviders)
    gpt_4o_mini: OpenAIModelConfig = OpenAIModelConfig(providers=OpenAiProviders)
    llama_3_70b: Llama370BModelConfig = Llama370BModelConfig(
        tools_providers=Llama370bToolsProviders,
        no_tools_providers=Llama370bNoToolsProviders,
    )
    claude_3_5_sonnet: AnthropicModelConfig = AnthropicModelConfig(
        providers=AnthropicProviders
    )
    claude_3_haiku: AnthropicModelConfig = AnthropicModelConfig(
        providers=AnthropicProviders
    )
    gemini_1_5_pro: GoogleModelConfig = GoogleModelConfig(providers=GoogleProviders)
    gemini_1_5_flash: GoogleModelConfig = GoogleModelConfig(providers=GoogleProviders)
    command_r_plus: CohereModelConfig = CohereModelConfig(providers=CohereProviders)
    command_r: CohereModelConfig = CohereModelConfig(providers=CohereProviders)

    @classmethod
    def OpenAI(cls, providers: list[Provider] = OpenAiProviders) -> OpenAIModelConfig:
        return OpenAIModelConfig(providers=providers)

    @classmethod
    def Anthropic(
        cls, providers: list[Provider] = AnthropicProviders
    ) -> AnthropicModelConfig:
        return AnthropicModelConfig(providers=providers)

    @classmethod
    def Google(cls, providers: list[Provider] = GoogleProviders) -> GoogleModelConfig:
        return GoogleModelConfig(providers=providers)

    @classmethod
    def Cohere(cls, providers: list[Provider] = CohereProviders) -> CohereModelConfig:
        return CohereModelConfig(providers=providers)

    @classmethod
    def Llama370B(
        cls,
        tools_providers: list[Provider] = Llama370bToolsProviders,
        no_tools_providers: list[Provider] = Llama370bNoToolsProviders,
    ) -> Llama370BModelConfig:
        return Llama370BModelConfig(
            tools_providers=tools_providers, no_tools_providers=no_tools_providers
        )


class FallbackConfig(BaseModel):
    # By default just fall back through models recommended by the router from best to worst.
    fallback_model: Optional[LLM] = None

    # By default don't fall back at all and only try the top model recommended by the router.
    max_model_fallback_attempts: int = 0

    # By default use all available providers until a successful response is received.
    max_provider_fallback_attempts: int = sys.maxsize


class ToolsConfig(BaseModel):
    # By default assume no parallel tool use
    parallel_tool_use: bool = False


class Dials(BaseModel):
    quality: float = 1
    cost: float = 0

    def sum_to_one(self) -> bool:
        return (self.quality + self.cost) == 1


class DialtoneClient(BaseModel):
    api_key: str
    provider_config: ProviderConfig
    dials: Dials = Dials()
    router_model_config: RouterModelConfig = RouterModelConfig()
    fallback_config: FallbackConfig = FallbackConfig()
    tools_config: ToolsConfig = ToolsConfig()
    base_url: str = DEFAULT_BASE_URL


class RouteDecision(BaseModel):
    model: LLM
    providers: list[Provider]
    quality_predictions: dict[str, float]
    routing_strategy: str
